import random
from collections import defaultdict

from sourced.ml.algorithms.uast_ids_to_bag import FakeVocabulary, Uast2BagBase
from sourced.ml.algorithms import uast2sequence


class Uast2StructBagBase(Uast2BagBase):
    SEP = ">"

    def __init__(self, stride, seq_len, node2index=None):
        self._node2index = node2index if node2index is not None else FakeVocabulary()
        self._stride = stride
        if not isinstance(seq_len, (int, tuple, list)):
            raise TypeError("Unexpected type of seq_len: %s" % type(seq_len))
        self._seq_lens = [seq_len] if isinstance(seq_len, int) else seq_len

    @property
    def node2index(self):
        return self._node2index


class Node2InternalType:
    # FIXME(zurk): change to simple function. Vadim Markovtsev comments:
    # > would rather made this a simple function and change roles2index
    # type from [] to callable. Saves time to understand.
    def __getitem__(self, item):
        return item.internal_type


class UastSeq2Bag(Uast2StructBagBase):
    """
    DFS traversal + preserves the order of node children.
    """

    def __init__(self, stride=1, seq_len=(3, 4), node2index=None):
        _node2index = Node2InternalType() if node2index is None else node2index
        super().__init__(stride, seq_len, _node2index)

    def __call__(self, uast):
        bag = defaultdict(int)
        node_sequence = uast2sequence(uast)

        # convert to str - requirement from wmhash.BagsExtractor
        node_sequence = [self.node2index[n] for n in node_sequence]

        for seq_len in self._seq_lens:
            for i in range(0, len(node_sequence) - seq_len + 1, self._stride):
                key = self.SEP.join(node_sequence[i:i + seq_len])
                bag[key] += 1
        return bag


class Node:
    def __init__(self, parent=None, internal_type=None):
        self.parent = parent
        self.internal_type = internal_type
        self.children = []

    @property
    def neighbours(self):
        neighbours = []
        if self.parent is not None:
            neighbours.append(self.parent)
        neighbours.extend(self.children)
        return neighbours


class Uast2RandomWalks:
    """
    Generation of random walks for UAST.
    """

    def __init__(self, p_explore_neighborhood, q_leave_neighborhood, n_walks, n_steps,
                 node2index=None, seed=None):
        """
        Related article: https://arxiv.org/abs/1607.00653

        :param p_explore_neighborhood: return parameter, p. Parameter p controls the likelihood of\
                                       immediately revisiting a node in the walk. Setting it to a\
                                       high value (> max(q, 1)) ensures that we are less likely to\
                                       sample an already visited node in the following two steps\
                                       (unless the next node in the walk had no other neighbor).\
                                       This strategy encourages moderate exploration and avoids\
                                       2-hop redundancy in sampling.
        :param q_leave_neighborhood: in-out parameter, q. Parameter q allows the search to\
                                     differentiate between “inward” and “outward” nodes. Such \
                                     walks obtain a local view of the underlying graph with \
                                     respect to the start node in the walk and approximate BFS \
                                     behavior in the sense that our samples comprise of nodes \
                                     within a small locality.
        :param n_walks: Number of walks from each node.
        :param n_steps: Number of steps in walk.
        """
        self.p_explore_neighborhood = p_explore_neighborhood
        self.q_leave_neighborhood = q_leave_neighborhood
        self.n_walks = n_walks
        self.n_steps = n_steps
        self.node2index = node2index if node2index is not None else Node2InternalType()
        if seed is not None:
            random.seed(seed)

    def __call__(self, uast):
        starting_nodes = self.prepare_starting_nodes(uast)
        for i in range(self.n_walks):
            for start_node in starting_nodes:
                yield self.random_walk(start_node)

    @staticmethod
    def _extract_node(node, parent):
        return Node(parent=parent, internal_type=node.internal_type)

    def prepare_starting_nodes(self, uast):
        starting_nodes = []
        root = self._extract_node(uast, None)
        stack = [(root, uast)]
        while stack:
            parent, parent_uast = stack.pop()
            children_nodes = [self._extract_node(child, parent) for child in parent_uast.children]
            parent.children = children_nodes
            stack.extend(zip(children_nodes, parent_uast.children))
            starting_nodes.append(parent)

        return starting_nodes

    def random_walk(self, node):
        walk = [node]
        while len(walk) < self.n_steps:
            walk.append(self.alias_sample(walk))

        walk = [self.node2index[n] for n in walk]
        return walk

    def alias_sample(self, walk):
        """
        Compare to node2vec this sampling is a bit simpler because there is no loop in tree ->
        so there are only 2 options with unnormalized probabilities 1/p & 1/q
        Related article: https://arxiv.org/abs/1607.00653

        :param walk: list of visited nodes
        :return: next node to visit
        """
        last_node = walk[-1]  # correspond to node v in article

        if len(walk) == 1:
            choice_list = last_node.children
            if last_node.parent is not None:
                choice_list.append(last_node.parent)
            if len(choice_list) == 0:
                return last_node
            return random.choice(last_node.children)

        threshold = (1 / self.p_explore_neighborhood)
        threshold /= (threshold + len(last_node.children) / self.q_leave_neighborhood)

        if random.random() <= threshold:
            # With threshold probability we need to return back to previous node.
            return walk[-2]  # Node from previous step. Correspond to node t in article.

        return random.choice(last_node.neighbours)


class UastRandomWalk2Bag(Uast2StructBagBase):
    def __init__(self, p_explore_neighborhood=0.79, q_leave_neighborhood=0.82, n_walks=2,
                 n_steps=10, stride=1, seq_len=(2, 3), seed=42):
        super().__init__(stride, seq_len)
        self.uast2walks = Uast2RandomWalks(p_explore_neighborhood=p_explore_neighborhood,
                                           q_leave_neighborhood=q_leave_neighborhood,
                                           n_walks=n_walks, n_steps=n_steps, seed=seed)

    def __call__(self, uast):
        bag = defaultdict(int)
        for walk in self.uast2walks(uast):
            for seq_len in self._seq_lens:
                for i in range(0, len(walk) - seq_len + 1, self._stride):
                    # convert to str - requirement from wmhash.BagsExtractor
                    bag[self.SEP.join(walk[i:i + seq_len])] += 1
        return bag



import logging

from sourced.ml.models import MergeDocFreq
from sourced.ml.cmd.args import handle_input_arg


def merge_df(args):
    log = logging.getLogger("merge_df")
    merger = MergeDocFreq(ordered=args.ordered,
                          vocabulary_size=args.vocabulary_size,
                          min_docfreq=args.min_docfreq)
    merger.convert(handle_input_arg(args.input, log), args.output)



from pyspark import Row, RDD, SparkContext

from sourced.ml.algorithms import log_tf_log_idf
from sourced.ml.transformers.uast2bag_features import Uast2BagFeatures
from sourced.ml.transformers.transformer import Transformer


class TFIDF(Transformer):
    """
    Calculates TF-IDF (log-log) weight for every feature.
    """
    Columns = Uast2BagFeatures.Columns

    def __init__(self, df: dict, ndocs: int, sc: SparkContext, **kwargs):
        """
        :param df: dict containing document frequencies calculated for the given stream.
        :param ndocs: total number of documents
        :param sc: spark context used to broadcast `df
        """
        super().__init__(**kwargs)
        self.df = df
        self.ndocs = ndocs
        self.sc = sc

    def __call__(self, head: RDD):
        """

        :param head: pyspark rdd where each row is named tuple with `token`, `document` and `value`
                   (term frequency) fields. One can use Uast2TermFreq Transformer to calculate
                   such rdd.
        :return: rdd after applying TFIDF
        """
        c = self.Columns
        df = self.sc.broadcast(self.df)
        ndocs = self.ndocs
        head = head \
            .filter(lambda x: df.value.get(x[c.token]) is not None) \
            .map(lambda x: Row(**{
                c.token: x[c.token],
                c.document: x[c.document],
                c.value: log_tf_log_idf(df=df.value[x[c.token]], tf=x[c.value], ndocs=ndocs)}))
        df.unpersist(blocking=True)
        return head



import logging
from typing import Iterable, List, Dict

from scipy import sparse

from modelforge import Model, split_strings, assemble_sparse_matrix, \
    merge_strings, disassemble_sparse_matrix, register_model
from modelforge.progress_bar import progress_bar
from sourced.ml.models.df import DocumentFrequencies


@register_model
class BOW(Model):
    """
    Weighted bag of words model. Every word is correspond to an index and its matrix column.
    Bag is a word set from repository, file or anything else.
    Word is source code identifier or its part.
    This model depends on :class:`sourced.ml.models.DocumentFrequencies`.
    """
    NAME = "bow"

    def construct(self, documents: List[str], tokens: List[str], matrix: sparse.spmatrix):
        if matrix.shape[0] != len(documents):
            raise ValueError("matrix shape mismatch, documents %d != %d" % (
                matrix.shape[0], len(documents)))
        if matrix.shape[1] != len(tokens):
            raise ValueError("matrix shape mismatch, tokens %d != %d" % (
                matrix.shape[1], len(tokens)))
        self._documents = documents
        self._matrix = matrix
        self._tokens = tokens
        return self

    def dump(self):
        return "Shape: %s\n" \
               "First 10 documents: %s\n" \
               "First 10 tokens: %s" % \
               (self._matrix.shape, self._documents[:10], self.tokens[:10])

    @property
    def matrix(self) -> sparse.spmatrix:
        """
        Returns the bags as a sparse matrix. Rows are documents and columns are tokens weight.
        """
        return self._matrix

    @property
    def documents(self):
        """
        The list of documents in the model.
        """
        return self._documents

    @property
    def tokens(self):
        """
        The list of tokens in the model.
        """
        return self._tokens

    def __getitem__(self, item: int):
        """
        Returns document name, word indices and weights for the given document index.

        :param item: Document index.
        :return: (name, :class:`numpy.ndarray` with word indices, \
                  :class:`numpy.ndarray` with weights)
        """
        data = self._matrix[item]
        return self._documents[item], data.indices, data.data

    def __iter__(self):
        """
        Returns an iterator over the document indices.
        """
        return iter(range(len(self)))

    def __len__(self):
        """
        Returns the number of documents.
        """
        return len(self._documents)

    def save(self, output: str, deps: Iterable=tuple()):
        if not deps:
            try:
                deps = [self.get_dep(DocumentFrequencies.NAME)]
            except KeyError:
                raise ValueError(
                    "You must specify DocumentFrequencies dependency to save BOW.") from None
        super().save(output, deps)

    def convert_bow_to_vw(self, output: str):
        log = logging.getLogger("bow2vw")
        log.info("Writing %s", output)
        with open(output, "w") as fout:
            for index in progress_bar(self, log, expected_size=len(self)):
                record = self[index]
                fout.write(record[0].replace(":", "").replace(" ", "_") + " ")
                pairs = []
                for t, v in zip(*record[1:]):
                    try:
                        word = self.tokens[t]
                    except (KeyError, IndexError):
                        log.warning("%d not found in the vocabulary", t)
                        continue
                    pairs.append("%s:%s" % (word, v))
                fout.write(" ".join(pairs))
                fout.write("\n")

    def documents_index(self) -> Dict[str, int]:
        return {r: i for i, r in enumerate(self._documents)}

    def _generate_tree(self):
        return {"documents": merge_strings(self._documents),
                "matrix": disassemble_sparse_matrix(self._matrix),
                "tokens": merge_strings(self.tokens)}

    def _load_tree_kwargs(self, tree: dict):
        return dict(documents=split_strings(tree["documents"]),
                    matrix=assemble_sparse_matrix(tree["matrix"]),
                    tokens=split_strings(tree["tokens"]))

    def _load_tree(self, tree: dict):
        self.construct(**self._load_tree_kwargs(tree))



import functools
import logging
import requests
from pkg_resources import get_distribution, DistributionNotFound
from sourced.engine import Engine
from sourced.ml.utils import add_spark_args, create_spark, SparkDefault


def get_engine_version():
    try:
        engine = get_distribution("sourced-engine").version
    except DistributionNotFound:
        log = logging.getLogger("engine_version")
        engine = requests.get("https://api.github.com/repos/src-d/engine/releases/latest") \
            .json()["tag_name"].replace("v", "")
        log.warning("Engine not found, queried GitHub to get the latest release tag (%s)",
                    engine)
    return engine


class EngineDefault:
    """
    Engine default initialization parameters
    """
    BBLFSH = "localhost"
    VERSION = get_engine_version()
    REPOSITORY_FORMAT = "siva"


class EngineConstants:
    """
    Constants for Engine usage.
    """
    class Columns:
        """
        Column names constants.
        """
        RepositoryId = "repository_id"
        Path = "path"
        BlobId = "blob_id"
        Uast = "uast"


def add_engine_args(my_parser, default_packages=None):
    add_spark_args(my_parser, default_packages=default_packages)
    my_parser.add_argument("--bblfsh", default=EngineDefault.BBLFSH,
                           help="Babelfish server's address.")
    my_parser.add_argument("--engine", default=EngineDefault.VERSION,
                           help="source{d} engine version.")
    my_parser.add_argument("--repository-format", default=EngineDefault.REPOSITORY_FORMAT,
                           help="Repository storage input format.")


def get_engine_package(engine):
    return "tech.sourced:engine:" + engine


def get_bblfsh_dependency(bblfsh):
    return "spark.tech.sourced.bblfsh.grpc.host=" + bblfsh


def create_engine(session_name, repositories,
                  repository_format=EngineDefault.REPOSITORY_FORMAT,
                  bblfsh=EngineDefault.BBLFSH,
                  engine=EngineDefault.VERSION,
                  config=SparkDefault.CONFIG,
                  packages=SparkDefault.JAR_PACKAGES,
                  spark=SparkDefault.MASTER_ADDRESS,
                  spark_local_dir=SparkDefault.LOCAL_DIR,
                  spark_log_level=SparkDefault.LOG_LEVEL,
                  dep_zip=SparkDefault.DEP_ZIP,
                  memory=SparkDefault.MEMORY):

    config += (get_bblfsh_dependency(bblfsh),)
    packages += (get_engine_package(engine),)
    session = create_spark(session_name, spark=spark, spark_local_dir=spark_local_dir,
                           config=config, packages=packages, spark_log_level=spark_log_level,
                           dep_zip=dep_zip, memory=memory)
    logging.getLogger("engine").info("Initializing engine on %s", repositories)
    return Engine(session, repositories, repository_format)


def pause(func):
    @functools.wraps(func)
    def wrapped_pause(cmdline_args, *args, **kwargs):
        try:
            return func(cmdline_args, *args, **kwargs)
        finally:
            if cmdline_args.pause:
                input("Press Enter to exit...")

    return wrapped_pause


def pipeline_graph(args, log, root):
    if args.graph:
        log.info("Dumping the graph to %s", args.graph)
        with open(args.graph, "w") as f:
            root.graph(stream=f)



class Foo:
    def func_a(self):
        # should be counted
        pass


def func_b():
    # should be counted
    pass


def func_c():
    # should be counted
    def func_d():
        # should not be counted
        pass



from collections import defaultdict
import os

from sourced.ml.models.model_converters.base import Model2Base
from sourced.ml.models.df import DocumentFrequencies
from sourced.ml.models.ordered_df import OrderedDocumentFrequencies


class MergeDocFreq(Model2Base):
    """
    Merges several :class:`DocumentFrequencies` models together.
    """
    MODEL_FROM_CLASS = DocumentFrequencies
    MODEL_TO_CLASS = DocumentFrequencies

    def __init__(self, min_docfreq: int, vocabulary_size: int, ordered: bool=False,
                 *args, **kwargs):
        super().__init__(num_processes=1, *args, **kwargs)
        self.ordered = ordered
        self.min_docfreq = min_docfreq
        self.vocabulary_size = vocabulary_size
        self._df = defaultdict(int)
        self._docs = 0

    def convert_model(self, model: DocumentFrequencies) -> None:
        for word, freq in model:
            self._df[word] += freq
        self._docs += model.docs

    def finalize(self, index: int, destdir: str):
        df_model = OrderedDocumentFrequencies if self.ordered else DocumentFrequencies
        df_model(log_level=self._log.level) \
            .construct(self._docs, self._df) \
            .prune(self.min_docfreq) \
            .greatest(self.vocabulary_size) \
            .save(self._save_path(index, destdir))

    @staticmethod
    def _save_path(index: int, destdir: str):
        if destdir.endswith(".asdf"):
            return destdir
        return os.path.join(destdir, "docfreq_%d.asdf" % index)



import logging
from typing import Iterable, Tuple

from sourced.ml.algorithms import Uast2QuantizedChildren
from sourced.ml.extractors import BagsExtractor, register_extractor, filter_kwargs, \
    get_names_from_kwargs


@register_extractor
class ChildrenBagExtractor(BagsExtractor):
    """
    Converts a UAST to the bag of pairs (internal type, quantized number of children).
    """
    NAME = "children"
    NAMESPACE = "c."
    OPTS = dict(get_names_from_kwargs(Uast2QuantizedChildren.__init__))

    def __init__(self, docfreq_threshold=None, **kwargs):
        original_kwargs = kwargs
        uast2bag_kwargs = filter_kwargs(kwargs, Uast2QuantizedChildren.__init__)
        for k in uast2bag_kwargs:
            kwargs.pop(k)
        super().__init__(docfreq_threshold, **kwargs)
        self._log.debug("__init__ %s", original_kwargs)
        self.uast_to_bag = Uast2QuantizedChildren(**uast2bag_kwargs)

    @property
    def npartitions(self):
        return self.uast_to_bag.npartitions

    @property
    def levels(self):
        return self.uast_to_bag.levels

    def extract(self, uast):
        if not self.uast_to_bag.levels:
            # bypass NAMESPACE
            gen = self.uast_to_bag(uast).items()
        else:
            gen = super().extract(uast)
        for key, val in gen:
            yield key, val

    def quantize(self, frequencies: Iterable[Tuple[str, Iterable[Tuple[int, int]]]]):
        self.uast_to_bag.quantize(frequencies)
        if self._log.isEnabledFor(logging.DEBUG):
            for k, v in self.uast_to_bag.levels.items():
                self._log.debug("%s\n%s", k, v)



# flake8: noqa
from sourced.ml.extractors.helpers import __extractors__, get_names_from_kwargs, \
    register_extractor, filter_kwargs, create_extractors_from_args
from sourced.ml.extractors.bags_extractor import Extractor, BagsExtractor, RoleIdsExtractor
from sourced.ml.extractors.identifiers import IdentifiersBagExtractor
from sourced.ml.extractors.literals import LiteralsBagExtractor
from sourced.ml.extractors.uast_random_walk import UastRandomWalkBagExtractor
from sourced.ml.extractors.uast_seq import UastSeqBagExtractor
from sourced.ml.extractors.children import ChildrenBagExtractor
from sourced.ml.extractors.graphlets import GraphletBagExtractor
from sourced.ml.extractors.identifier_distance import IdentifierDistance
from sourced.ml.extractors.id_sequence import IdSequenceExtractor



import logging
import os
import shutil
import warnings

import numpy
try:
    import tensorflow as tf
except ImportError as e:
    warnings.warn("Tensorflow is not installed, dependent functionality is unavailable.")

from modelforge.progress_bar import progress_bar
from sourced.ml.algorithms.id_embedding import extract_coocc_matrix
from sourced.ml.models import Cooccurrences, DocumentFrequencies


def _int64s(xs):
    return tf.train.Feature(
        int64_list=tf.train.Int64List(value=list(xs)))


def _floats(xs):
    return tf.train.Feature(
        float_list=tf.train.FloatList(value=list(xs)))


def id2vec_preprocess(args):
    """
    Loads co-occurrence matrices for several repositories and generates the
    document frequencies and the Swivel protobuf dataset.

    :param args: :class:`argparse.Namespace` with "input", "vocabulary_size", \
                 "shard_size", "df" and "output".
    :return: None
    """
    log = logging.getLogger("preproc")
    log.info("Loading docfreq model from %s", args.docfreq_in)
    df_model = DocumentFrequencies(log_level=args.log_level).load(source=args.docfreq_in)
    coocc_model = Cooccurrences().load(args.input)
    if numpy.any(coocc_model.matrix.data < 0):
        raise ValueError(("Co-occurrence matrix %s contains negative elements. "
                          "Please check its correctness.") % args.input)
    if numpy.any(numpy.isnan(coocc_model.matrix.data)):
        raise ValueError(("Co-occurrence matrix %s contains nan elements. "
                          "Please check its correctness.") % args.input)

    try:
        df_meta = coocc_model.get_dep(DocumentFrequencies.NAME)
        if df_model.meta != df_meta:
            raise ValueError((
                "Document frequency model you provided does not match dependency inside "
                "Cooccurrences model:\nargs.docfreq.meta:\n%s\ncoocc_model.get_dep"
                "(\"docfreq\")\n%s\n") % (df_model.meta, df_meta))
    except KeyError:
        pass  # There is no docfreq dependency

    vs = args.vocabulary_size
    if len(df_model) < vs:
        vs = len(df_model)
    sz = args.shard_size
    if vs < sz:
        raise ValueError(
            "vocabulary_size=%s is less than shard_size=%s. You should specify a smaller "
            "shard_size (e.g. shard_size=%s)." % (vs, sz, vs))
    vs -= vs % sz
    log.info("Effective vocabulary size: %d", vs)
    df_model = df_model.greatest(vs)
    log.info("Sorting the vocabulary...")
    chosen_words = sorted(df_model.tokens())
    word_indices = {w: i for i, w in enumerate(chosen_words)}

    if not os.path.exists(args.output):
        os.makedirs(args.output)
    with open(os.path.join(args.output, "row_vocab.txt"), "w") as out:
        out.write('\n'.join(chosen_words))
    log.info("Saved row_vocab.txt")
    shutil.copyfile(os.path.join(args.output, "row_vocab.txt"),
                    os.path.join(args.output, "col_vocab.txt"))
    log.info("Saved col_vocab.txt")
    del chosen_words

    ccmatrix = extract_coocc_matrix((vs, vs), word_indices, coocc_model)

    log.info("Planning the sharding...")
    bool_sums = ccmatrix.indptr[1:] - ccmatrix.indptr[:-1]
    reorder = numpy.argsort(-bool_sums)
    with open(os.path.join(args.output, "row_sums.txt"), "w") as out:
        out.write('\n'.join(map(str, bool_sums.tolist())))
    log.info("Saved row_sums.txt")
    shutil.copyfile(os.path.join(args.output, "row_sums.txt"),
                    os.path.join(args.output, "col_sums.txt"))
    log.info("Saved col_sums.txt")

    log.info("Writing the shards...")
    os.makedirs(args.output, exist_ok=True)
    nshards = vs // args.shard_size
    for row in progress_bar(range(nshards), log, expected_size=nshards):
        for col in range(nshards):
            indices_row = reorder[row::nshards]
            indices_col = reorder[col::nshards]
            shard = ccmatrix[indices_row][:, indices_col].tocoo()

            example = tf.train.Example(features=tf.train.Features(feature={
                "global_row": _int64s(indices_row),
                "global_col": _int64s(indices_col),
                "sparse_local_row": _int64s(shard.row),
                "sparse_local_col": _int64s(shard.col),
                "sparse_value": _floats(shard.data)}))

            with open(os.path.join(args.output,
                                   "shard-%03d-%03d.pb" % (row, col)),
                      "wb") as out:
                out.write(example.SerializeToString())
    log.info("Success")



from collections import defaultdict

from sourced.ml.algorithms.uast_ids_to_bag import Uast2BagBase
from sourced.ml.algorithms.uast_struct_to_bag import Node


class Uast2GraphletBag(Uast2BagBase):
    """
    Converts a UAST to a bag of graphlets.
    The graphlet of a UAST node is composed from the node itself, its parent and its children.
    Each node is represented by the internal role string.
    """
    @staticmethod
    def _extract_node(node, parent):
        return Node(parent=parent, internal_type=node.internal_type)

    def uast2graphlets(self, uast):
        """
        :param uast: The UAST root node.
        :generate: The nodes which compose the UAST.
            :class: 'Node' is used to access the nodes of the graphlets.
        """
        root = self._extract_node(uast, None)
        stack = [(root, uast)]
        while stack:
            parent, parent_uast = stack.pop()
            children_nodes = [self._extract_node(child, parent) for child in parent_uast.children]
            parent.children = children_nodes
            stack.extend(zip(children_nodes, parent_uast.children))
            yield parent

    def node2key(self, node):
        """
        Builds the string joining internal types of all the nodes
        in the node's graphlet in the following order:
        parent_node_child1_child2_child3. The children are sorted by alphabetic order.
        str format is required for BagsExtractor.

        :param node: a node of UAST
        :return: The string key of node
        """
        try:
            parent_type = node.parent.internal_type
        except AttributeError:
            parent_type = None
        key = [parent_type, node.internal_type]
        key.extend(sorted(ch.internal_type for ch in node.children))
        return "_".join(map(str, key))

    def __call__(self, uast):
        """
        Converts a UAST to a weighed bag of graphlets. The weights are graphlets frequencies.
        :param uast: The UAST root node.
        :return: bag of graphlets.
        """
        bag = defaultdict(int)
        for node in self.uast2graphlets(uast):
            bag[self.node2key(node)] += 1
        return bag



from typing import Union

from modelforge import register_model, Model, split_strings, assemble_sparse_matrix, \
    merge_strings, disassemble_sparse_matrix


@register_model
class Topics(Model):
    NAME = "topics"

    @property
    def tokens(self):
        return self._tokens

    @property
    def topics(self):
        """
        May be None if no topics are labeled.
        """
        return self._topics

    @property
    def matrix(self):
        """
        Rows: tokens
        Columns: topics
        """
        return self._matrix

    def construct(self, tokens: list, topics: Union[list, None], matrix):
        if len(tokens) != matrix.shape[1]:
            raise ValueError("Tokens and matrix do not match.")
        self._tokens = tokens
        self._topics = topics
        self._matrix = matrix
        return self

    def _load_tree(self, tree: dict) -> None:
        self.construct(split_strings(tree["tokens"]),
                       split_strings(tree["topics"]) if tree["topics"] else None,
                       assemble_sparse_matrix(tree["matrix"]))

    def dump(self) -> str:
        res = "%d topics, %d tokens\nFirst 10 tokens: %s\nTopics: " % (
            self.matrix.shape + (self.tokens[:10],))
        if self.topics is not None:
            res += "labeled, first 10: %s\n" % self.topics[:10]
        else:
            res += "unlabeled\n"
        nnz = self.matrix.getnnz()
        res += "non-zero elements: %d  (%f)" % (
            nnz, nnz / (self.matrix.shape[0] * self.matrix.shape[1]))
        return res

    def _generate_tree(self):
        return {"tokens": merge_strings(self.tokens),
                "topics": merge_strings(self.topics) if self.topics is not None else False,
                "matrix": disassemble_sparse_matrix(self.matrix)}

    def __len__(self):
        """
        Returns the number of topics.
        """
        return self.matrix.shape[0]

    def __getitem__(self, item):
        """
        Returns the keywords sorted by significance from topic index.
        """
        row = self.matrix[item]
        nnz = row.nonzero()[1]
        pairs = [(-row[0, i], i) for i in nnz]
        pairs.sort()
        return [(self.tokens[pair[1]], -pair[0]) for pair in pairs]

    def label_topics(self, labels):
        if len(labels) != len(self):
            raise ValueError("Sizes do not match: %d != %d" % (len(labels), len(self)))
        if not isinstance(labels[0], str):
            raise TypeError("Labels must be strings")
        self._topics = list(labels)



import numpy


def extract_coocc_matrix(global_shape, word_indices, model):
    # Stage 1 - extract the tokens, map them to the global vocabulary
    indices = []
    mapped_indices = []
    for i, w in enumerate(model.tokens):
        gi = word_indices.get(w)
        if gi is not None:
            indices.append(i)
            mapped_indices.append(gi)
    indices = numpy.array(indices)
    mapped_indices = numpy.array(mapped_indices)
    # Stage 2 - sort the matched tokens by the index in the vocabulary
    order = numpy.argsort(mapped_indices)
    indices = indices[order]
    mapped_indices = mapped_indices[order]
    # Stage 3 - produce the csr_matrix with the matched tokens **only**
    matrix = model.matrix.tocsr()[indices][:, indices]
    # Stage 4 - convert this matrix to the global (ccmatrix) coordinates
    csr_indices = matrix.indices
    for i, v in enumerate(csr_indices):
        # Here we use the fact that indices and mapped_indices are in the same order
        csr_indices[i] = mapped_indices[v]
    csr_indptr = matrix.indptr
    new_indptr = [0]
    for i, v in enumerate(mapped_indices):
        prev_ptr = csr_indptr[i]
        ptr = csr_indptr[i + 1]

        # Handle missing rows
        prev = (mapped_indices[i - 1] + 1) if i > 0 else 0
        for z in range(prev, v):
            new_indptr.append(prev_ptr)

        new_indptr.append(ptr)
    for z in range(mapped_indices[-1] + 1, global_shape[0]):
        new_indptr.append(csr_indptr[-1])
    matrix.indptr = numpy.array(new_indptr)
    matrix._shape = global_shape
    return matrix



# flake8: noqa
from sourced.ml.algorithms.tf_idf import log_tf_log_idf
from sourced.ml.algorithms.token_parser import TokenParser, NoopTokenParser
from sourced.ml.algorithms.uast_ids_to_bag import UastIds2Bag, uast2sequence
from sourced.ml.algorithms.uast_struct_to_bag import UastRandomWalk2Bag, UastSeq2Bag
from sourced.ml.algorithms.uast_inttypes_to_nodes import Uast2QuantizedChildren
from sourced.ml.algorithms.uast_inttypes_to_graphlets import Uast2GraphletBag
from sourced.ml.algorithms.uast_to_role_id_pairs import Uast2RoleIdPairs
from sourced.ml.algorithms.uast_id_distance import Uast2IdLineDistance, Uast2IdTreeDistance
from sourced.ml.algorithms.uast_to_id_sequence import Uast2IdSequence



import logging
from uuid import uuid4

from sourced.ml.extractors import IdSequenceExtractor
from sourced.ml.transformers import UastDeserializer, Uast2BagFeatures, UastRow2Document, \
    CsvSaver, create_uast_source
from sourced.ml.transformers.basic import Rower
from sourced.ml.utils.engine import pipeline_graph, pause


@pause
def repos2id_sequence(args):
    log = logging.getLogger("repos2id_distance")
    extractor = IdSequenceExtractor(args.split)
    session_name = "repos2roles_and_ids-%s" % uuid4()
    root, start_point = create_uast_source(args, session_name)
    if not args.skip_docname:
        mapper = Rower(lambda x: dict(document=x[0][1],
                                      identifiers=x[0][0]))
    else:
        mapper = Rower(lambda x: dict(identifiers=x[0][0]))
    start_point \
        .link(UastRow2Document()) \
        .link(UastDeserializer()) \
        .link(Uast2BagFeatures(extractor)) \
        .link(mapper) \
        .link(CsvSaver(args.output)) \
        .execute()
    pipeline_graph(args, log, root)



import os
from scipy.sparse import vstack

from sourced.ml import extractors
from sourced.ml.models.bow import BOW
from sourced.ml.models.model_converters.base import Model2Base


class MergeBOW(Model2Base):
    """
    Merges several :class:`BOW` models together.
    """
    MODEL_FROM_CLASS = BOW
    MODEL_TO_CLASS = BOW

    def __init__(self, features=None, *args, **kwargs):
        super().__init__(num_processes=1, *args, **kwargs)
        self.documents = None
        self.tokens = None
        self.matrix = None
        self.deps = None
        self.features_namespaces = None
        if features:
            self.features_namespaces = [ex.NAMESPACE for ex in extractors.__extractors__.values()
                                        if ex.NAME in features]

    def convert_model(self, model: BOW) -> None:
        if self.tokens is None:
            self.tokens = model.tokens
            self.documents = model.documents
            self.matrix = [model.matrix.tocsr()]
            self.deps = model._meta["dependencies"]
        elif set(self.tokens) != set(model.tokens):
            raise ValueError("Models don't share the same set of tokens !")
        else:
            self.documents += model.documents
            self.matrix.append(model.matrix.tocsr())

    def finalize(self, index: int, destdir: str):
        self._log.info("Stacking matrices ...")
        matrix = self.matrix.pop(0)
        while self.matrix:
            matrix = vstack([matrix, self.matrix.pop(0)])
            self._log.info("%s matrices to stack ...", len(self.matrix))
        self.matrix = matrix
        self._log.info("Writing model ...")
        if self.features_namespaces:
            self._reduce_matrix()
        BOW(log_level=self._log.level) \
            .construct(self.documents, self.tokens, self.matrix) \
            .save(self._save_path(index, destdir), self.deps)

    def _reduce_matrix(self):
        reduced_tokens = []
        columns = []
        matrix = self.matrix.tocsc()
        for i, token in enumerate(self.tokens):
            if token.split(".")[0] in self.features_namespaces:
                reduced_tokens.append(token)
                columns.append(i)
        self.tokens = reduced_tokens
        self.matrix = matrix[:, columns]

    @staticmethod
    def _save_path(index: int, destdir: str):
        if destdir.endswith(".asdf"):
            return destdir
        return os.path.join(destdir, "bow_%d.asdf" % index)



import os
import logging
from typing import List

from sourced.ml.models import QuantizationLevels
from sourced.ml.transformers import Uast2Quant
from sourced.ml.extractors import BagsExtractor


def create_or_apply_quant(model_path: str, extractors: List[BagsExtractor], extracted_uasts=None):
    log = logging.getLogger("create_or_apply_quant")
    if os.path.exists(model_path):
        log.info("Loading the quantization levels from %s and applying quantization to supported"
                 " extractors...", model_path)
        try:
            QuantizationLevels().load(source=model_path).apply_quantization(extractors)
        except (ValueError, ImportError):
            pass
        else:
            return
    if extracted_uasts is None:
        log.error("[IN] only mode, please supply a quantization levels model")
        raise ValueError
    else:
        quant = Uast2Quant(extractors)
        extracted_uasts.link(quant).execute()
        if quant.levels:
            log.info("Writing quantization levels to %s", model_path)
            QuantizationLevels().construct(quant.levels).save(model_path)



import glob
import logging
import os

import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split, KFold, GridSearchCV
from tqdm import tqdm

from sourced.ml.cmd.args import handle_input_arg
from sourced.ml.models import Id2Vec


def load_dataset(directory):
    csvpaths = glob.glob(os.path.join(directory, "**/*.csv"))
    chunks = []
    for csvpath in tqdm(csvpaths):
        chunks.append(pd.read_csv(csvpath, sep=',', index_col=None))
    df = pd.concat(chunks)
    df.role = df.role.map(lambda x: x.replace("IDENTIFIER | ", "").replace(" | IDENTIFIER", ""))
    return df


def identifiers_to_datasets(df_unique, id2vecs, log):
    """
    Replace identifiers with its embeddings and create standard X, y dataset
    """
    y = df_unique["role"]
    identifiers = df_unique["identifier"]
    log.info("Final dataset size is %d" % len(identifiers))
    Xs = {}
    for name, id2vec in id2vecs.items():
        X = []
        for key in identifiers:
            X.append(id2vec.embeddings[id2vec[key]])
        Xs[name] = np.array(X)

    return Xs, y


def get_quality(X, y, estimator, tuned_parameters, seed, log):
    X_train, X_test, y_train, y_test = \
        train_test_split(X, y, test_size=0.33, random_state=seed)

    folding = KFold(n_splits=5, shuffle=True, random_state=seed)
    clf = GridSearchCV(estimator, tuned_parameters, cv=folding, n_jobs=-1)
    clf.fit(X_train, y_train)

    log.debug("Best parameters set found on development set:", clf.best_params_)
    log.debug("Grid scores on development set:")
    means = clf.cv_results_['mean_test_score']
    stds = clf.cv_results_['std_test_score']
    best_values = (0,)
    for mean, std, params in zip(means, stds, clf.cv_results_['params']):
        log.debug("\t%0.3f (+/-%0.03f) for %r" % (mean, std, params))
        if best_values[0] < mean:
            best_values = (mean, std, params)
    return clf.score(X_test, y_test), params


def id2role_eval(args):
    """
    Build a simple log-regression model to predict a Role of UAST node by its identifier embedding.
    It creates a report about embedding quality.
    To collect the dataset please use repos2roleids entry point.
    """
    log = logging.getLogger("id2role_eval")

    models = {}
    common_tokens = None
    for path in handle_input_arg(args.models, log):
        name = os.path.split(path)[1]
        id2vec = Id2Vec().load(path)
        id2vec.construct(id2vec.embeddings, [t.split(".")[-1] for t in id2vec.tokens])
        models[name] = id2vec
        if common_tokens is None:
            common_tokens = set(models[name].tokens)
        else:
            common_tokens &= set(models[name].tokens)
    log.info("Common tokens in all models: %d" % len(common_tokens))

    tuned_parameters = [{'C': [10 ** x for x in range(-7, -1)]}]
    # Load data and preprocess
    log.info("Data loading...")
    df = load_dataset(args.dataset)
    df_ids = set(df["identifier"])
    valid_tokens = list(set(df_ids) & common_tokens)
    df = df[df["identifier"].isin(valid_tokens)]
    # Count identifiers in dataset
    log.info("Have embeddings only for %d tokens from %d in your dataset" % (
        len(valid_tokens), len(df_ids)))
    df_unique = df.groupby("identifier").agg(lambda x: x.value_counts().index[0])
    df_unique["identifier"] = df_unique.index
    # Exclude rare roles
    vc = df["role"].value_counts()
    del df
    rare = set(vc[vc < 10].index)
    log.info("%d rare roles excluded. " % len(rare))
    df_unique = df_unique[~df_unique["role"].isin(rare)]
    log.debug("Convert words to its embeddings")
    Xs, y = identifiers_to_datasets(df_unique, models, log)

    final_report = pd.DataFrame(columns=["embedding name", "score", "best C value"])
    for name in tqdm(Xs):
        log.info("{}...".format(name))
        best_values = get_quality(Xs[name], y,
                                  LogisticRegression(class_weight="balanced",
                                                     random_state=args.seed),
                                  tuned_parameters=tuned_parameters,
                                  seed=args.seed,
                                  log=log)
        final_report = final_report.append({"embedding name": name,
                                            "score": best_values[0],
                                            "best C value": best_values[1]["C"]},
                                           ignore_index=True)

    print("Pairs number: %d.\n" % len(valid_tokens))
    print(final_report)



from typing import Union, Iterable, Tuple

from bblfsh import Node
import numpy

from sourced.ml.algorithms.uast_to_bag import Uast2BagThroughSingleScan


class Uast2QuantizedChildren(Uast2BagThroughSingleScan):
    """
    Converts a UAST to a bag of children counts.
    """

    def __init__(self, npartitions: int=20):
        self.npartitions = npartitions
        self.levels = {}

    def node2key(self, node: Node) -> Union[str, Tuple[str, int]]:
        """
        :param node: a node in UAST.
        :return: The string which consists of the internal type of the node and its number of
        children.
        """
        if not self.levels:
            return node.internal_type, len(node.children)
        qm = self.levels[node.internal_type]
        quant_index = numpy.searchsorted(qm, len(node.children), side="right") - 1
        return "%s_%d" % (node.internal_type, quant_index)

    def quantize(self, frequencies: Iterable[Tuple[str, Iterable[Tuple[int, int]]]]):
        for key, vals in frequencies:
            self.levels[key] = self.quantize_unwrapped(vals)

    def quantize_unwrapped(self, children_freq: Iterable[Tuple[int, int]]) -> numpy.ndarray:
        """
        Builds the quantization partition P that is a vector of length nb_partitions \
        whose entries are in strictly ascending order.
        Quantization of x is defined as:
            0 if x <= P[0]
            m if P[m-1] < x <= P[m]
            n if P[n] <= x

        :param children_freq: distribution of the number of children.
        :return: The array with quantization levels.
        """
        levels = numpy.zeros(self.npartitions + 1, dtype=numpy.int32)
        children_freq = sorted(children_freq)
        max_nodes_per_bin = sum(i[1] for i in children_freq) / self.npartitions
        levels[0] = children_freq[0][0]
        accum = children_freq[0][1]
        i = 1
        for v, f in children_freq[1:]:
            accum += f
            if accum > max_nodes_per_bin:
                accum = f
                if i < len(levels):
                    levels[i] = v
                    i += 1
        last = children_freq[-1][0]
        if i < len(levels):
            levels[i:] = last
        else:
            levels[-1] = last
        return levels



import bblfsh


IDENTIFIER = bblfsh.role_id("IDENTIFIER")
QUALIFIED = bblfsh.role_id("QUALIFIED")
LITERAL = bblfsh.role_id("LITERAL")
OPERATOR = bblfsh.role_id("OPERATOR")
EXPRESSION = bblfsh.role_id("EXPRESSION")
LEFT = bblfsh.role_id("LEFT")
BINARY = bblfsh.role_id("BINARY")
ASSIGNMENT = bblfsh.role_id("ASSIGNMENT")
FUNCTION = bblfsh.role_id("FUNCTION")
DECLARATION = bblfsh.role_id("DECLARATION")
NAME = bblfsh.role_id("NAME")



import logging

import numpy

from sourced.ml.models import Id2Vec


def id2vec_project(args):
    MAX_TOKENS = 10000  # hardcoded in Tensorflow Projector

    log = logging.getLogger("id2vec_project")
    id2vec = Id2Vec(log_level=args.log_level).load(source=args.input)
    if args.docfreq_in:
        log.info("Loading docfreq model from %s", args.docfreq_in)
        from sourced.ml.models import DocumentFrequencies
        df = DocumentFrequencies(log_level=args.log_level).load(source=args.docfreq_in)
    else:
        df = None
    if len(id2vec) < MAX_TOKENS:
        tokens = numpy.arange(len(id2vec), dtype=int)
        if df is not None:
            freqs = [df.get(id2vec.tokens[i], 0) for i in tokens]
        else:
            freqs = None
    else:
        if df is not None:
            log.info("Filtering tokens through docfreq")
            items = []
            for token, idx in id2vec.items():
                try:
                    items.append((df[token], idx))
                except KeyError:
                    continue
            log.info("Sorting")
            items.sort(reverse=True)
            tokens = [i[1] for i in items[:MAX_TOKENS]]
            freqs = [i[0] for i in items[:MAX_TOKENS]]
        else:
            log.warning("You have not specified --df => picking random %d tokens", MAX_TOKENS)
            numpy.random.seed(777)
            tokens = numpy.random.choice(
                numpy.arange(len(id2vec), dtype=int), MAX_TOKENS, replace=False)
            freqs = None
    log.info("Gathering the embeddings")
    embeddings = numpy.vstack([id2vec.embeddings[i] for i in tokens])
    tokens = [id2vec.tokens[i] for i in tokens]
    labels = ["subtoken"]
    if freqs is not None:
        labels.append("docfreq")
        tokens = list(zip(tokens, (str(i) for i in freqs)))
    import sourced.ml.utils.projector as projector
    projector.present_embeddings(args.output, not args.no_browser, labels, tokens, embeddings)
    if not args.no_browser:
        projector.wait()



from bblfsh.client import BblfshClient
from distutils.version import StrictVersion

BBLFSH_VERSION_LOW = "2.2"
BBLFSH_VERSION_HIGH = "3.0"


def check_version(host: str="0.0.0.0", port: str="9432") -> bool:
    """
    Check if the bblfsh server version matches module requirements.

    :param host: bblfsh server host
    :param port: bblfsh server port
    :return: True if bblfsh version specified matches requirements
    """
    # get version and remove leading 'v'
    version = StrictVersion(BblfshClient("%s:%s" % (host, port)).version().version.lstrip("v"))
    return StrictVersion(BBLFSH_VERSION_LOW) <= version < StrictVersion(BBLFSH_VERSION_HIGH)



import os


VENDOR = "source{d}"
BACKEND = "gcs"
BACKEND_ARGS = "bucket=models.cdn.sourced.tech"
INDEX_REPO = "https://github.com/src-d/models"
CACHE_DIR = os.path.join(os.path.expanduser("~"), ".cache", "source{d}")



import argparse
import logging
import os
import sys

from modelforge.logs import setup_logging

from sourced.ml import extractors
from sourced.ml.transformers import Moder
from sourced.ml import cmd
from sourced.ml.cmd import args
from sourced.ml.cmd.run_swivel import mirror_tf_args
from sourced.ml.utils import install_bigartm, add_spark_args


def get_parser() -> argparse.ArgumentParser:
    """
    Creates the cmdline argument parser.
    """

    parser = argparse.ArgumentParser(formatter_class=args.ArgumentDefaultsHelpFormatterNoNone)
    parser.add_argument("--log-level", default="INFO",
                        choices=logging._nameToLevel,
                        help="Logging verbosity.")
    # Create and construct subparsers
    subparsers = parser.add_subparsers(help="Commands", dest="command")

    def add_parser(name, help_message):
        return subparsers.add_parser(
            name, help=help_message, formatter_class=args.ArgumentDefaultsHelpFormatterNoNone)

    # ------------------------------------------------------------------------
    preprocessing_parser = subparsers.add_parser(
        "preprocrepos", help="Convert siva to parquet files with extracted information.")
    preprocessing_parser.set_defaults(handler=cmd.preprocess_repos)
    preprocessing_parser.add_argument("-x", "--mode", choices=Moder.Options.__all__,
                                      default="file", help="What to extract from repositories.")
    args.add_repo2_args(preprocessing_parser)
    preprocessing_parser.add_argument(
        "-o", "--output", required=True,
        help="[OUT] Path to the parquet files with bag batches.")
    default_fields = ("blob_id", "repository_id", "content", "path", "commit_hash", "uast", "lang")
    preprocessing_parser.add_argument(
        "-f", "--fields", nargs="+", default=default_fields,
        help="Fields to save.")
    # ------------------------------------------------------------------------
    repos2bow_parser = add_parser(
        "repos2bow", "Convert source code to the bag-of-words model.")
    repos2bow_parser.set_defaults(handler=cmd.repos2bow)
    args.add_df_args(repos2bow_parser)
    args.add_repo2_args(repos2bow_parser)
    args.add_feature_args(repos2bow_parser)
    args.add_bow_args(repos2bow_parser)
    args.add_repartitioner_arg(repos2bow_parser)
    args.add_cached_index_arg(repos2bow_parser)
    # ------------------------------------------------------------------------
    repos2bow_index_parser = add_parser(
        "repos2bow_index", "Creates the index, quant and docfreq model of the bag-of-words model.")
    repos2bow_index_parser.set_defaults(handler=cmd.repos2bow_index)
    args.add_df_args(repos2bow_index_parser)
    args.add_repo2_args(repos2bow_index_parser)
    args.add_feature_args(repos2bow_index_parser)
    args.add_repartitioner_arg(repos2bow_index_parser)
    args.add_cached_index_arg(repos2bow_index_parser, create=True)
    # ------------------------------------------------------------------------
    repos2df_parser = add_parser(
        "repos2df", "Calculate document frequencies of features extracted from source code.")
    repos2df_parser.set_defaults(handler=cmd.repos2df)
    args.add_df_args(repos2df_parser)
    args.add_repo2_args(repos2df_parser)
    args.add_feature_args(repos2df_parser)
    # ------------------------------------------------------------------------
    repos2ids_parser = subparsers.add_parser(
        "repos2ids", help="Convert source code to a bag of identifiers.")
    repos2ids_parser.set_defaults(handler=cmd.repos2ids)
    args.add_repo2_args(repos2ids_parser)
    args.add_split_stem_arg(repos2ids_parser)
    args.add_repartitioner_arg(repos2ids_parser)
    repos2ids_parser.add_argument(
        "-o", "--output", required=True,
        help="[OUT] output path to the CSV file with identifiers.")
    repos2ids_parser.add_argument(
        "--idfreq", action="store_true",
        help="Adds identifier frequencies to the output CSV file."
             "num_repos is the number of repositories where the identifier appears in."
             "num_files is the number of files where the identifier appears in."
             "num_occ is the total number of occurences of the identifier.")
    # ------------------------------------------------------------------------
    repos2coocc_parser = add_parser(
        "repos2coocc", "Convert source code to the sparse co-occurrence matrix of identifiers.")
    repos2coocc_parser.set_defaults(handler=cmd.repos2coocc)
    args.add_df_args(repos2coocc_parser)
    args.add_repo2_args(repos2coocc_parser)
    args.add_split_stem_arg(repos2coocc_parser)
    args.add_repartitioner_arg(repos2coocc_parser)
    repos2coocc_parser.add_argument(
        "-o", "--output", required=True,
        help="[OUT] Path to the Cooccurrences model.")

    # ------------------------------------------------------------------------
    repos2roles_and_ids = add_parser(
        "repos2roleids", "Converts a UAST to a list of pairs, where pair is a role and "
        "identifier. Role is merged generic roles where identifier was found.")
    repos2roles_and_ids.set_defaults(handler=cmd.repos2roles_and_ids)
    args.add_repo2_args(repos2roles_and_ids)
    args.add_split_stem_arg(repos2roles_and_ids)
    repos2roles_and_ids.add_argument(
        "-o", "--output", required=True,
        help="[OUT] Path to the directory where spark should store the result. "
             "Inside the direcory you find result is csv format, status file and sumcheck files.")
    # ------------------------------------------------------------------------
    repos2identifier_distance = add_parser(
        "repos2id_distance", "Converts a UAST to a list of identifier pairs "
                             "and distance between them.")
    repos2identifier_distance.set_defaults(handler=cmd.repos2id_distance)
    args.add_repo2_args(repos2identifier_distance)
    args.add_split_stem_arg(repos2identifier_distance)
    repos2identifier_distance.add_argument(
        "-t", "--type", required=True, choices=extractors.IdentifierDistance.DistanceType.All,
        help="Distance type.")
    repos2identifier_distance.add_argument(
        "--max-distance", default=extractors.IdentifierDistance.DEFAULT_MAX_DISTANCE, type=int,
        help="Maximum distance to save.")
    repos2identifier_distance.add_argument(
        "-o", "--output", required=True,
        help="[OUT] Path to the directory where spark should store the result. "
             "Inside the direcory you find result is csv format, status file and sumcheck files.")
    # ------------------------------------------------------------------------
    repos2id_sequence = add_parser(
        "repos2idseq", "Converts a UAST to sequence of identifiers sorted by order of appearance.")
    repos2id_sequence.set_defaults(handler=cmd.repos2id_sequence)
    args.add_repo2_args(repos2id_sequence)
    args.add_split_stem_arg(repos2id_sequence)
    repos2id_sequence.add_argument(
        "--skip-docname", default=False, action="store_true",
        help="Do not save document name in CSV file, only identifier sequence.")
    repos2id_sequence.add_argument(
        "-o", "--output", required=True,
        help="[OUT] Path to the directory where spark should store the result. "
             "Inside the direcory you find result is csv format, status file and sumcheck files.")
    # ------------------------------------------------------------------------
    preproc_parser = add_parser(
        "id2vec-preproc", "Convert a sparse co-occurrence matrix to the Swivel shards.")
    preproc_parser.set_defaults(handler=cmd.id2vec_preprocess)
    args.add_df_args(preproc_parser)
    preproc_parser.add_argument("-s", "--shard-size", default=4096, type=int,
                                help="The shard (submatrix) size.")
    preproc_parser.add_argument(
        "-i", "--input",
        help="Concurrence model produced by repos2coocc.")
    preproc_parser.add_argument("-o", "--output", required=True, help="Output directory.")
    # ------------------------------------------------------------------------
    train_parser = add_parser(
        "id2vec-train", "Train identifier embeddings using Swivel.")
    train_parser.set_defaults(handler=cmd.run_swivel)
    mirror_tf_args(train_parser)
    # ------------------------------------------------------------------------
    id2vec_postproc_parser = add_parser(
        "id2vec-postproc",
        "Combine row and column embeddings produced by Swivel and write them to an .asdf.")
    id2vec_postproc_parser.set_defaults(handler=cmd.id2vec_postprocess)
    id2vec_postproc_parser.add_argument(
        "-i", "--swivel-data", required=True,
        help="Folder with swivel row and column embeddings data. "
             "You can get it using id2vec_train subcommand.")
    id2vec_postproc_parser.add_argument(
        "-o", "--output", required=True,
        help="Output directory for Id2Vec model.")
    # ------------------------------------------------------------------------
    id2vec_project_parser = add_parser(
        "id2vec-project", "Present id2vec model in Tensorflow Projector.")
    id2vec_project_parser.set_defaults(handler=cmd.id2vec_project)
    args.add_df_args(id2vec_project_parser, required=False)
    id2vec_project_parser.add_argument("-i", "--input", required=True,
                                       help="id2vec model to present.")
    id2vec_project_parser.add_argument("-o", "--output", required=True,
                                       help="Projector output directory.")
    id2vec_project_parser.add_argument("--no-browser", action="store_true",
                                       help="Do not open the browser.")
    # ------------------------------------------------------------------------
    train_id_split_parser = add_parser(
        "train-id-split", "Train a neural network to split identifiers.")
    train_id_split_parser.set_defaults(handler=cmd.train_id_split)
    # common arguments for CNN/RNN models
    train_id_split_parser.add_argument("-i", "--input", required=True,
                                       help="Path to the input data in CSV format:"
                                            "num_files,num_occ,num_repos,token,token_split")
    train_id_split_parser.add_argument("-e", "--epochs", type=int, default=10,
                                       help="Number of training epochs. The more the better"
                                            "but the training time is proportional.")
    train_id_split_parser.add_argument("-b", "--batch-size", type=int, default=500,
                                       help="Batch size. Higher values better utilize GPUs"
                                            "but may harm the convergence.")
    train_id_split_parser.add_argument("-l", "--length", type=int, default=40,
                                       help="RNN sequence length.")
    train_id_split_parser.add_argument("-o", "--output", required=True,
                                       help="Path to store the trained model.")
    train_id_split_parser.add_argument("-t", "--test-ratio", type=float, default=0.2,
                                       help="Fraction of the dataset to use for evaluation.")
    train_id_split_parser.add_argument("-p", "--padding", default="post", choices=("pre", "post"),
                                       help="Wether to pad before or after each sequence.")
    train_id_split_parser.add_argument("--optimizer", default="Adam", choices=("RMSprop", "Adam"),
                                       help="Algorithm to use as an optimizer for the neural net.")
    train_id_split_parser.add_argument("--lr", default=0.001, type=float,
                                       help="Initial learning rate.")
    train_id_split_parser.add_argument("--final-lr", default=0.00001, type=float,
                                       help="Final learning rate. The decrease from "
                                            "the initial learning rate is done linearly.")
    train_id_split_parser.add_argument("--samples-before-report", type=int, default=5*10**6,
                                       help="Number of samples between each validation report"
                                            "and training updates.")
    train_id_split_parser.add_argument("--val-batch-size", type=int, default=2000,
                                       help="Batch size for validation."
                                            "It can be increased to speed up the pipeline but"
                                            "it proportionally increases the memory consumption.")
    train_id_split_parser.add_argument("--seed", type=int, default=1989,
                                       help="Random seed.")
    train_id_split_parser.add_argument("--devices", default="0",
                                       help="Device(s) to use. '-1' means CPU.")
    train_id_split_parser.add_argument("--csv-identifier", default=3,
                                       help="Column name in the CSV file for the raw identifier.")
    train_id_split_parser.add_argument("--csv-identifier-split", default=4,
                                       help="Column name in the CSV file for the splitted"
                                            "identifier.")
    train_id_split_parser.add_argument("--include-csv-header", action="store_true",
                                       help="Treat the first line of the input CSV as a regular"
                                            "line.")
    train_id_split_parser.add_argument("--model", type=str, choices=("RNN", "CNN"), required=True,
                                       help="Neural Network model to use to learn the identifier"
                                            "splitting task.")
    train_id_split_parser.add_argument("-s", "--stack", default=2, type=int,
                                       help="Number of layers stacked on each other.")
    # RNN specific arguments
    train_id_split_parser.add_argument("--type-cell", default="LSTM",
                                       choices=("GRU", "LSTM", "CuDNNLSTM", "CuDNNGRU"),
                                       help="Recurrent layer type to use.")
    train_id_split_parser.add_argument("-n", "--neurons", default=256, type=int,
                                       help="Number of neurons on each layer.")
    # CNN specific arguments
    train_id_split_parser.add_argument("-f", "--filters", default="64,32,16,8",
                                       help="Number of filters for each kernel size.")
    train_id_split_parser.add_argument("-k", "--kernel-sizes", default="2,4,8,16",
                                       help="Sizes for sliding windows.")
    train_id_split_parser.add_argument("--dim-reduction", default=32, type=int,
                                       help="Number of 1-d kernels to reduce dimensionality"
                                            "after each layer.")
    # ------------------------------------------------------------------------
    bow2vw_parser = add_parser(
        "bow2vw", "Convert a bag-of-words model to the dataset in Vowpal Wabbit format.")
    bow2vw_parser.set_defaults(handler=cmd.bow2vw)
    bow2vw_parser.add_argument(
        "--bow", help="URL or path to a bag-of-words model.")
    bow2vw_parser.add_argument(
        "--id2vec", help="URL or path to the identifier embeddings.")
    bow2vw_parser.add_argument(
        "-o", "--output", required=True, help="Path to the output file.")
    # ------------------------------------------------------------------------
    bigartm_postproc_parser = add_parser(
        "bigartm2asdf", "Convert a human-readable BigARTM model to Modelforge format.")
    bigartm_postproc_parser.set_defaults(handler=cmd.bigartm2asdf)
    bigartm_postproc_parser.add_argument("input")
    bigartm_postproc_parser.add_argument("output")
    # ------------------------------------------------------------------------
    bigartm_parser = add_parser(
        "bigartm", "Install bigartm/bigartm to the current working directory.")
    bigartm_parser.set_defaults(handler=install_bigartm)
    bigartm_parser.add_argument(
        "--tmpdir", help="Store intermediate files in this directory instead of /tmp.")
    bigartm_parser.add_argument("--output", default=os.getcwd(), help="Output directory.")

    # ------------------------------------------------------------------------
    merge_df = add_parser("merge-df", "Merge DocumentFrequencies models to a single one.")
    merge_df.set_defaults(handler=cmd.merge_df)
    args.add_min_docfreq(merge_df)
    args.add_vocabulary_size_arg(merge_df)
    merge_df.add_argument(
        "-o", "--output", required=True,
        help="Path to the merged DocumentFrequencies model.")
    merge_df.add_argument(
        "-i", "--input", required=True, nargs="+",
        help="DocumentFrequencies models input files."
             "Use `-i -` to read input files from stdin.")
    merge_df.add_argument(
        "--ordered", action="store_true", default=False,
        help="Save OrderedDocumentFrequencies. "
             "If not specified DocumentFrequencies model will be saved")
    # ------------------------------------------------------------------------
    merge_coocc = add_parser("merge-coocc", "Merge several Cooccurrences models together.")
    merge_coocc.set_defaults(handler=cmd.merge_coocc)
    add_spark_args(merge_coocc)
    merge_coocc.add_argument(
        "-o", "--output", required=True,
        help="Path to the merged Cooccurrences model.")
    merge_coocc.add_argument(
        "-i", "--input", required=True,
        help="Cooccurrences models input files."
             "Use `-i -` to read input files from stdin.")
    merge_coocc.add_argument(
        "--docfreq", required=True,
        help="[IN] Specify OrderedDocumentFrequencies model. "
             "Identifiers that are not present in the model will be ignored.")
    merge_coocc.add_argument(
        "--no-spark", action="store_true", default=False,
        help="Use the local reduction instead of PySpark. "
             "Can be faster and consume less memory if the data fits into RAM.")
    # ------------------------------------------------------------------------
    merge_bow = add_parser("merge-bow", "Merge BOW models to a single one.")
    merge_bow.set_defaults(handler=cmd.merge_bow)
    merge_bow.add_argument(
        "-i", "--input", required=True, nargs="+",
        help="BOW models input files."
             "Use `-i -` to read input files from stdin.")
    merge_bow.add_argument(
        "-o", "--output", required=True,
        help="Path to the merged BOW model.")
    merge_bow.add_argument(
        "-f", "--features", nargs="+",
        choices=[ex.NAME for ex in extractors.__extractors__.values()],
        default=None, help="To keep only specific features, if not specified all will be kept.")
    # ------------------------------------------------------------------------
    id2role_eval = add_parser("id2role-eval",
                              "Compare the embeddings quality on role prediction problem.")
    id2role_eval.set_defaults(handler=cmd.id2role_eval)
    id2role_eval.add_argument(
        "-m", "--models", required=True, nargs="+",
        help="Id2Vec models to compare."
             "Use `-i -` to read input files from stdin.")
    id2role_eval.add_argument(
        "-d", "--dataset", required=True,
        help="Dataset directory. You can collect dataset via repos2roleids command.")
    id2role_eval.add_argument(
        "-s", "--seed", default=420,
        help="Random seed for reproducible results.")
    return parser


def main():
    """
    Creates all the argparse-rs and invokes the function from set_defaults().
    :return: The result of the function from set_defaults().
    """

    parser = get_parser()
    args = parser.parse_args()
    args.log_level = logging._nameToLevel[args.log_level]
    setup_logging(args.log_level)
    try:
        handler = args.handler
    except AttributeError:
        def print_usage(_):
            parser.print_usage()

        handler = print_usage
    return handler(args)


if __name__ == "__main__":
    sys.exit(main())



from importlib.machinery import SourceFileLoader
from setuptools import setup, find_packages
import sys

sourcedml = SourceFileLoader("sourced.ml", "./sourced/ml/__init__.py").load_module()

if sys.version_info < (3, 5, 0):
    typing = ["typing"]
else:
    typing = []

setup(
    name="sourced-ml",
    description="Framework for machine learning on source code. "
                "Provides API and tools to train and use models based "
                "on source code features extracted from Babelfish's UASTs.",
    version=".".join(map(str, sourcedml.__version__)),
    license="Apache 2.0",
    author="source{d}",
    author_email="machine-learning@sourced.tech",
    url="https://github.com/src-d/ml",
    download_url="https://github.com/src-d/ml",
    packages=find_packages(exclude=("sourced.ml.tests",)),
    namespace_packages=["sourced"],
    entry_points={
        "console_scripts": ["srcml=sourced.ml.__main__:main"],
    },
    keywords=["machine learning on source code", "word2vec", "id2vec",
              "github", "swivel", "bow", "bblfsh", "babelfish"],
    install_requires=["PyStemmer>=1.3,<2.0",
                      "bblfsh>=2.2.1,<3.0",
                      "modelforge>=0.7.0,<0.8",
                      "sourced-engine>=0.5.1,<0.7",
                      "humanize>=0.5.0",
                      "parquet>=1.2,<2.0",
                      "pygments>=2.2.0,<3.0",
                      "keras>=2.0,<3.0"] + typing,
    extras_require={
        "tf": ["tensorflow>=1.0,<2.0"],
        "tf_gpu": ["tensorflow-gpu>=1.0,<2.0"],
    },
    tests_require=["docker>=3.4.0,<4.0"],
    package_data={"": ["LICENSE.md", "README.md"],
                  "sourced": ["ml/transformers/languages.yml"], },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Libraries"
    ]
)



import logging


class PickleableLogger:
    """
    Base class which provides the logging features through ``self._log``.
    Can be safely pickled.
    """
    def __init__(self, log_level=logging.INFO):
        self._log = logging.getLogger(self._get_log_name())
        self._log.setLevel(log_level)

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_log"] = self._log.level
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        log_level = state["_log"]
        self._log = logging.getLogger(self._get_log_name())
        self._log.setLevel(log_level)

    def _get_log_name(self):
        """
        Children must implement this method. It shall return the logger's name.
        """
        raise NotImplementedError



import argparse
import logging
from typing import List, Union

from sourced.engine.engine import BlobsDataFrame, BlobsWithLanguageDataFrame
from pyspark import RDD, Row, StorageLevel
from pyspark.sql import DataFrame, functions

from sourced.ml.extractors.helpers import filter_kwargs
from sourced.ml.transformers.transformer import Transformer
from sourced.ml.transformers.uast2bag_features import Uast2BagFeatures
from sourced.ml.utils import EngineConstants, get_spark_memory_config, create_engine, \
    create_spark, SparkDefault


class Repartitioner(Transformer):
    """
    Repartitioner uses one of the three ways to split an RDD into partitions:
    1. coalesce() if shuffle=False, keymap=None
    2. repartition() if shuffle=True, keymap=None
    3. partitionBy() if keymap is not None
    """
    def __init__(self, partitions: int, shuffle: bool=False, keymap: callable=None, **kwargs):
        super().__init__(**kwargs)
        self.partitions = partitions
        self.shuffle = shuffle
        self.keymap = keymap

    def __call__(self, head: RDD):
        if self.keymap is None:
            return head.coalesce(self.partitions, self.shuffle)
        # partitionBy the key extracted using self.keymap
        try:
            # this checks if keymap is an identity
            probe = self.keymap("probe")
        except:  # noqa: E722
            probe = None
        if probe != "probe":
            head = head.map(lambda x: (self.keymap(x), x))
        return head \
            .partitionBy(self.partitions) \
            .map(lambda x: x[1])

    @staticmethod
    def maybe(partitions: Union[int, None], shuffle: bool=False, keymap: callable=None,
              multiplier: int=1):
        if partitions is not None:
            return Repartitioner(partitions * multiplier, shuffle=shuffle, keymap=keymap)
        else:
            return Identity()


class PartitionSelector(Transformer):
    """
    PartitionSelector return the partition by specific index.
    """
    def __init__(self, partition_index: int, **kwargs):
        super().__init__(**kwargs)
        self.partition_index = partition_index

    def __call__(self, head: RDD):
        index = self.partition_index

        def partition_filter(split_index, part):
            if split_index == index:
                for row in part:
                    yield row

        return head.mapPartitionsWithIndex(partition_filter, True)


class CsvSaver(Transformer):
    def __init__(self, output: str, **kwargs):
        super().__init__(**kwargs)
        self.output = output

    def __call__(self, head: RDD):
        self._log.info("Writing %s", self.output)
        return head.toDF() \
            .coalesce(1) \
            .write \
            .option("header", "true") \
            .mode("overwrite") \
            .csv(self.output)


class Rower(Transformer):
    def __init__(self, dicter, **kwargs):
        super().__init__(**kwargs)
        self.dicter = dicter

    def __call__(self, head: RDD):
        return head.map(lambda x: Row(**self.dicter(x)))


class Sampler(Transformer):
    """
    Wraps `sample()` function from pyspark Dataframe.
    """
    def __init__(self, with_replacement=False, fraction=0.05, seed=42, **kwargs):
        super().__init__(**kwargs)
        self.with_replacement = with_replacement
        self.fraction = fraction
        self.seed = seed

    def __call__(self, head: RDD):
        return head.sample(self.with_replacement, self.fraction, self.seed)


class Collector(Transformer):
    def __call__(self, head: RDD):
        return head.collect()


class Distinct(Transformer):
    def __call__(self, head: RDD):
        return head.distinct()


class First(Transformer):
    def __call__(self, head: RDD):
        return head.first()


class Identity(Transformer):
    def __call__(self, head: RDD):
        return head


class Cacher(Transformer):
    def __init__(self, persistence, **kwargs):
        super().__init__(**kwargs)
        self.persistence = getattr(StorageLevel, persistence)
        self.head = None
        self.trace = None

    def __getstate__(self):
        state = super().__getstate__()
        state["head"] = None
        state["trace"] = None
        return state

    def __call__(self, head: RDD):
        if self.head is None or self.trace != self.path():
            self.head = head.persist(self.persistence)
            self.trace = self.path()
        return self.head

    @staticmethod
    def maybe(persistence):
        if persistence is not None:
            return Cacher(persistence)
        else:
            return Identity()

    def unpersist(self):
        self.head.unpersist()


class Ignition(Transformer):
    """
    All pipelines start with this transformer - it returns all the repositories from the engine.
    """
    def __init__(self, engine, **kwargs):
        super().__init__(**kwargs)
        self.engine = engine

    def __getstate__(self):
        state = super().__getstate__()
        del state["engine"]
        return state

    def __call__(self, _) -> DataFrame:
        return self.engine.repositories


class RepositoriesFilter(Transformer):
    """
    Filters repositories by a regular expression over the identifiers
    (ex. "github.com/src-d/vecino" or "file:///tmp/vecino-yzv92l0i/repo").
    """
    def __init__(self, idre: str, **kwargs):
        super().__init__(**kwargs)
        self.idre = idre

    def __call__(self, repositories: DataFrame) -> DataFrame:
        return repositories.filter(repositories["id"].rlike(self.idre))


class DzhigurdaFiles(Transformer):
    def __init__(self, dzhigurda, **kwargs):
        super().__init__(**kwargs)
        self.dzhigurda = dzhigurda

    def __call__(self, repositories: DataFrame) -> DataFrame:
        head_ref = repositories.references.head_ref
        if self.dzhigurda < 0:
            # Use all available commits
            chosen = head_ref.all_reference_commits
        elif self.dzhigurda == 0:
            # Use only the first commit on a reference.
            # This case is completely the same with HeadFiles Transformer
            chosen = head_ref.commits
        else:
            commits = head_ref.all_reference_commits
            chosen = commits.filter(commits.index <= self.dzhigurda)
        return chosen.tree_entries.blobs


class HeadFiles(Transformer):
    def __call__(self, repositories: DataFrame) -> DataFrame:
        return repositories.references.head_ref.commits.tree_entries.blobs


class Counter(Transformer):
    def __init__(self, distinct=False, approximate=False, **kwargs):
        super().__init__(**kwargs)
        self.distinct = distinct
        self.approximate = approximate

    def __call__(self, head: RDD):
        if self.distinct and not self.approximate:
            head = head.distinct()
        if self.explained:
            self._log.info("toDebugString():\n%s", head.toDebugString().decode())
        if not self.approximate or not self.distinct:
            return head.count()
        return head.countApproxDistinct()


class LanguageExtractor(Transformer):
    def __call__(self, files: BlobsDataFrame) -> DataFrame:
        if not isinstance(files, BlobsDataFrame):
            raise TypeError("Argument type is not BlobsDataFrame. "
                            "Language extraction can not be performed.")
        return files \
            .dropDuplicates(("blob_id",)) \
            .filter("is_binary = 'false'") \
            .classify_languages()


class LanguageSelector(Transformer):
    def __init__(self, languages: List[str], blacklist=False, **kwargs):
        super().__init__(**kwargs)
        self.languages = languages
        self.blacklist = blacklist

    def __call__(self, files: BlobsWithLanguageDataFrame) -> DataFrame:
        return files[files.lang.isin(self.languages) != self.blacklist]

    @staticmethod
    def maybe(languages, blacklist):
        if languages is None:
            return Identity()
        return LanguageSelector(languages, blacklist)


class UastExtractor(Transformer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self, files: Union[BlobsDataFrame, BlobsWithLanguageDataFrame]) -> DataFrame:
        if not isinstance(files, (BlobsDataFrame, BlobsWithLanguageDataFrame)):
            raise TypeError("Argument type should be BlobsDataFrame or BlobsWithLanguageDataFrame,"
                            " got %s" % type(files).__name__)
        # if UAST is not extracted, returns an empty list that we filter out here
        return files.extract_uasts().where(functions.size(functions.col("uast")) > 0)


class FieldsSelector(Transformer):
    def __init__(self, fields: Union[list, tuple], **kwargs):
        super().__init__(**kwargs)
        self.fields = fields

    def __call__(self, rdd: RDD) -> RDD:
        def select_fields(row):
            return Row(**{f: getattr(row, f) for f in self.fields})
        res = rdd.map(select_fields)
        if self.explained:
            self._log.info("toDebugString():\n%s", res.toDebugString().decode())
        return res


class ParquetSaver(Transformer):
    def __init__(self, save_loc, **kwargs):
        super().__init__(**kwargs)
        self.save_loc = save_loc

    def __call__(self, rdd: RDD):
        if self.explained:
            self._log.info("toDebugString():\n%s", rdd.toDebugString().decode())
        rdd.toDF().write.parquet(self.save_loc)


class ParquetLoader(Transformer):
    def __init__(self, session, paths, **kwargs):
        super().__init__(**kwargs)
        self.session = session
        self.paths = paths

    def __getstate__(self):
        state = super().__getstate__()
        del state["session"]
        return state

    def __call__(self, _):
        if isinstance(self.paths, (list, tuple)):
            return self.session.read.parquet(*self.paths)
        if isinstance(self.paths, str):
            return self.session.read.parquet(self.paths)
        raise ValueError


class UastDeserializer(Transformer):
    def __setstate__(self, state):
        super().__setstate__(state)
        from bblfsh import Node
        self.parse_uast = Node.FromString

    def __call__(self, rows: RDD) -> RDD:
        return rows.flatMap(self.deserialize_uast)

    def deserialize_uast(self, row: Row):
        if EngineConstants.Columns.Uast not in row:
            return
        if not row[EngineConstants.Columns.Uast]:
            return
        row_dict = row.asDict()
        row_dict[EngineConstants.Columns.Uast] = []
        for i, uast in enumerate(row[EngineConstants.Columns.Uast]):
            try:
                row_dict[EngineConstants.Columns.Uast].append(self.parse_uast(uast))
            except:  # noqa
                self._log.error("\nBabelfish Error: Failed to parse uast for document %s for uast "
                                "#%s" % (row[Uast2BagFeatures.Columns.document], i))
        yield Row(**row_dict)


def create_parquet_loader(session_name, repositories,
                          config=SparkDefault.CONFIG,
                          packages=SparkDefault.JAR_PACKAGES,
                          spark=SparkDefault.MASTER_ADDRESS,
                          spark_local_dir=SparkDefault.LOCAL_DIR,
                          spark_log_level=SparkDefault.LOG_LEVEL,
                          memory=SparkDefault.MEMORY,
                          dep_zip=SparkDefault.DEP_ZIP):
    config += get_spark_memory_config(memory)
    session = create_spark(session_name, spark=spark, spark_local_dir=spark_local_dir,
                           config=config, packages=packages, spark_log_level=spark_log_level,
                           dep_zip=dep_zip)
    log = logging.getLogger("parquet")
    log.info("Initializing on %s", repositories)
    parquet = ParquetLoader(session, repositories)
    return parquet


def create_file_source(args: argparse.Namespace, session_name: str):
    if args.parquet:
        parquet_loader_args = filter_kwargs(args.__dict__, create_parquet_loader)
        root = create_parquet_loader(session_name, **parquet_loader_args)
        file_source = root.link(LanguageSelector.maybe(languages=args.languages,
                                                       blacklist=args.blacklist))
    else:
        engine_args = filter_kwargs(args.__dict__, create_engine)
        root = Ignition(create_engine(session_name, **engine_args), explain=args.explain)
        file_source = root.link(DzhigurdaFiles(args.dzhigurda))
        if args.languages is not None:
            file_source = file_source \
                .link(LanguageExtractor()) \
                .link(LanguageSelector(languages=args.languages, blacklist=args.blacklist))

    return root, file_source


def create_uast_source(args: argparse.Namespace, session_name: str):
    root, file_source = create_file_source(args, session_name)
    if args.parquet:
        # Assume that we already have uast column inside, because we cannot convert parquet files
        # back to sourced-engine format to extract UASTs.
        return root, file_source
    return root, file_source.link(UastExtractor())



from itertools import islice
from typing import Iterable, Union, Dict, List

import numpy

from modelforge import Model, split_strings, merge_strings, register_model


@register_model
class DocumentFrequencies(Model):
    """
    Document frequencies - number of times a source code identifier appeared
    in different repositories. Each repository counts only once.
    """
    NAME = "docfreq"

    def construct(self, docs: int, tokfreqs: Union[Iterable[Dict[str, int]], Dict[str, int]]):
        """
        Initializes this model.
        :param docs: The number of documents.
        :param tokfreqs: The dictionary of token -> frequency or the iterable collection of such
                         dictionaries.
        :return: self
        """
        if isinstance(tokfreqs, dict):
            df = tokfreqs
        else:
            df = {}
            for d in tokfreqs:
                df.update(d)
        self._docs = docs
        self._df = df
        return self

    """
    WE DO NOT ADD THIS

    def df(self) -> dict:
    """

    def _load_tree(self, tree: dict, tokens=None):
        if tokens is None:
            tokens = split_strings(tree["tokens"])
        freqs = tree["freqs"]
        self._log.info("Building the docfreq dictionary...")
        tokfreq = dict(zip(tokens, freqs))
        self.construct(docs=tree["docs"], tokfreqs=tokfreq)

    def _generate_tree(self):
        tokens = self.tokens()
        freqs = numpy.array([self._df[t] for t in tokens], dtype=numpy.float32)
        return {"docs": self.docs, "tokens": merge_strings(tokens), "freqs": freqs}

    def dump(self):
        return """Number of words: %d
Random 10 words: %s
Number of documents: %d""" % (
            len(self._df), dict(islice(self._df.items(), 10)), self.docs)

    @property
    def docs(self) -> int:
        """
        Returns the number of documents.
        """
        return self._docs

    """
    WE DO NOT ADD THIS

    def df(self) -> dict:
    """

    def prune(self, threshold: int) -> "DocumentFrequencies":
        """
        Removes tokens which occur less than `threshold` times.
        The operation happens *not* in-place - a new model is returned.
        :param threshold: Minimum number of occurrences.
        :return: The new model if the current one had to be changed, otherwise self.
        """
        if threshold < 1:
            raise ValueError("Invalid threshold: %d" % threshold)
        if threshold == 1:
            return self
        self._log.info("Pruning to min %d occurrences", threshold)
        pruned = type(self)()
        pruned._docs = self.docs
        pruned._df = {k: v for k, v in self._df.items() if v >= threshold}
        self._log.info("Size: %d -> %d", len(self), len(pruned))
        pruned._meta = self.meta
        return pruned

    def greatest(self, max_size: int) -> "DocumentFrequencies":
        """
        Truncates the model to most frequent `max_size` tokens.
        The operation happens *not* in-place - a new model is returned.
        :param max_size: The maximum vocabulary size.
        :return: The new model if the current one had to be changed, otherwise self.
        """
        if max_size < 1:
            raise ValueError("Invalid max_size: %d" % max_size)
        if len(self) <= max_size:
            return self
        self._log.info("Pruning to max %d size", max_size)
        pruned = type(self)()
        pruned._docs = self.docs
        freqs = numpy.fromiter(self._df.values(), dtype=numpy.int32, count=len(self))
        keys = numpy.array(list(self._df.keys()), dtype=object)
        chosen = numpy.argpartition(freqs, len(freqs) - max_size)[len(freqs) - max_size:]
        border_freq = freqs[chosen].min()
        chosen = freqs >= border_freq
        # argpartition can leave some of the elements with freq == border_freq outside
        # so next step ensures that we include everything.
        freqs = freqs[chosen]
        keys = keys[chosen]
        # we need to be deterministic at the cutoff frequency
        # argpartition returns random samples every time
        # so we treat words with the cutoff frequency separately
        if max_size != freqs.shape[0]:
            assert max_size < freqs.shape[0]
            border_freq_indexes = freqs == border_freq
            border_keys = keys[border_freq_indexes]
            border_keys.sort()
            border_keys = border_keys[:max_size - freqs.shape[0]]
            df = dict(zip(keys[~border_freq_indexes], freqs[~border_freq_indexes]))
            df.update({key: border_freq for key in border_keys})
        else:
            df = dict(zip(keys, freqs))
        pruned._df = df
        self._log.info("Size: %d -> %d", len(self), len(pruned))
        pruned._meta = self.meta
        return pruned

    def __getitem__(self, item):
        return self._df[item]

    def __iter__(self):
        return iter(self._df.items())

    def __len__(self):
        """
        Returns the number of tokens in the model.
        """
        return len(self._df)

    def get(self, item, default=None) -> Union[int, None]:
        """
        Return the document frequency for a given token.

        :param item: The token to query.
        :param default: Returned value in case the token is missing.
        :return: int or `default`
        """
        return self._df.get(item, default)

    def tokens(self) -> List[str]:
        """
        Returns the list of tokens.
        """
        return list(self._df)

    """
    WE DO NOT ADD THIS

    def df(self) -> dict:
    """
